-- ////////////////// Helper -- /////////////////

local _M = {}

DATADOME_API_ENDPOINT = "api.datadome.co"

_M.DATADOME_API_CONFIG = {
  endpoint = DATADOME_API_ENDPOINT,
  ssl = true,
  port = 443,
  path = "/validate-request",
  timeout = 150
}

headersLength = {
  ['SecCHUAMobile']           = 8,
  ['SecCHDeviceMemory']       = 8,
  ['SecFetchUser']            = 8,
  ['SecCHUAArch']             = 16,
  ['SecCHUAPlatform']         = 32,
  ['SecFetchDest']            = 32,
  ['SecFetchMode']            = 32,
  ['SecFetchSite']            = 64,
  ['SecCHUA']                 = 128,
  ['SecCHUAModel']            = 128,
  ['AcceptCharset']           = 128,
  ['AcceptEncoding']          = 128,
  ['CacheControl']            = 128,
  ['ClientID']                = 128,
  ['Connection']              = 128,
  ['Pragma']                  = 128,
  ['X-Requested-With']        = 128,
  ['ContentType']             = 128,
  ['From']                    = 128,
  ['TrueClientIP']            = 128,
  ['X-Real-IP']               = 128,
  ['AcceptLanguage']          = 256,
  ['SecCHUAFullVersionList']  = 256,
  ['Via']                     = 256,
  ['XForwardedForIP']         = -512,
  ['Accept']                  = 512,
  ['HeadersList']             = 512,
  ['Host']                    = 512,
  ['Origin']                  = 512,
  ['ServerHostname']          = 512,
  ['ServerName']              = 512,
  ['UserAgent']               = 768,
  ['Referer']                 = 1024,
  ['Request']                 = 2048
}

DATADOME_SESSION_BY_HEADER_HEADER = 'x-datadome-clientid'

-- Truncate Header methods
function _M.truncateHeaders(headers)
  for k,v in pairs(headers) do
    if headersLength[k] ~= nil then
      if headersLength[k] > 0 then
        headers[k] = string.sub(v, 1, headersLength[k])
      else  -- backward truncation
        headers[k] = string.sub(v, headersLength[k])
      end
    end
  end
  return headers
end

--//
function _M.getCurrentMicroTime()
  -- we need time up to microseccconds, but at lua we can do up to seconds :( round it
  return tostring(os.time()) .. "000000"
  -- return string.format("%i%06i",txn.f:date(),txn.f:date_us())
end
--
function _M.apiConnectionState(txn)
  if (txn.f:http_first_req() == 0) then
      return "reuse"
  else
      return "new"
  end
end
function _M.getProtocol(txn)
  if (txn.f:ssl_fc() == 0) then
      return "http"
  else
      return "https"
  end
end
--//

local function createApiConfig(user_defined)
  for key, _ in pairs(_M.DATADOME_API_CONFIG) do
      if user_defined[key] ~= nil then
        _M.DATADOME_API_CONFIG[key] = user_defined[key]
      end
  end
  return _M.DATADOME_API_CONFIG
end

function _M.activate(config)
  DATADOME_API_CONFIG = createApiConfig(config)
  if config.endpoint ~= nil then
      DATADOME_API_ENDPOINT = config.endpoint
  end
  local api_protocol = DATADOME_API_CONFIG.ssl and 'https://' or 'http://'
  DATADOME_API_CONFIG.api_url = api_protocol .. DATADOME_API_ENDPOINT .. ':' .. DATADOME_API_CONFIG.port .. DATADOME_API_CONFIG.path
  DATADOME_API_CONFIG.api_key = DATADOME_API_KEY
  core.log(core.debug, "Datadome Activated called with config: ".. _M.dump(config))
  return DATADOME_API_CONFIG
end

local function charToHex(c)
	return string.format("%%%02X", string.byte(c))
end

local function urlencode(url)
   if url == nil then
	  return
  end
  url = url:gsub("\n", "\r\n")
  url = url:gsub("([^%w_%%%-%.~])", charToHex)
  url = url:gsub(" ", "+")
  return url
end
-- //

function _M.dump(o)
	if type(o) == 'table' then
	   local s = '{ '
	   for k,v in pairs(o) do
		  if type(k) ~= 'number' then k = '"'..k..'"' end
		  s = s .. '['..k..'] = ' .. _M.dump(v) .. ','
	   end
	   return s .. '} '
	else
	   return tostring(o)
	end
 end

function _M.stringify(params)
    if type(params) == "table" then
      local fields = {}
      for key,value in pairs(params) do
        if value ~= nil then -- this removed empty values from data
          local keyString = core.concat() --
          keyString:add(urlencode(tostring(key)))
          keyString:add('=')
          if type(value) == "table" then
            for _, v in ipairs(value) do
              table.insert(fields, keyString:dump() .. urlencode(tostring(v)))
            end
          else
            table.insert(fields, keyString:dump() .. urlencode(tostring(value)))
          end
        end
      end
      return table.concat(fields, '&')
    end
    return ''
end
-- //
function _M.split(s, delimiter)
  local result = {};
  for match in (s..delimiter):gmatch("(.-)"..delimiter) do
      table.insert(result, match);
  end
  return result;
end

-- //
function _M.getClientIdAndCookiesLength(requestCookie)
    local cookie = requestCookie or ""
    local len = string.len(cookie)
    local clientId = nil
    if len > 0 then
		  local st, regex = Regex.new("datadome=([^;$]+)", true);
		  st, list = regex:match(cookie)
		  if st then
			  clientId = list[2]
		  end
    end
    return clientId, len
end
-- //
function _M.buildKeyList(params)
    if type(params) == "table" then
      local headersName = {}
      for key,value in pairs(params) do
        table.insert(headersName, key);
      end
      return table.concat(headersName, ',')
    end
    return '';
end

function _M.handleClientId(txn, body, datadomeHeaders, clientId)
  if txn.f:req_fhdr(DATADOME_SESSION_BY_HEADER_HEADER) ~= nil then
    core.log(core.info, 'sessionByHeader')
    body['ClientID'] = string.sub(txn.sf:req_hdr(DATADOME_SESSION_BY_HEADER_HEADER), 1, 128)
    datadomeHeaders["X-DataDome-X-Set-Cookie"] = "true"
  else
    if clientId ~= nil then
      body['ClientID'] = string.sub(clientId, 1, 128)
    end
  end
  return body, datadomeHeaders
end

return _M
