local helpers = require("core/helpers")

DATADOME_API_CONFIG = helpers.activate({
  endpoint = "api.datadome.co",
   timeout = 300
})

DATADOME_MODULE_NAME  = 'Haproxy-LUA-http'
DATADOME_MODULE_VERSION  = '1.0.1'

-- Parameters used to pass info in the requests
DATADOME_REQUEST_HEADERS_PARAMETER          = "txn.dd.x_datadome_request_headers"
DATADOME_RESPONSE_HEADERS_PARAMETER         = "txn.dd.x_datadome_headers"
DATADOME_RESPONSE_CODE_PARAMETER            = "txn.dd.x_datadome_response"
DATADOME_RESPONSE_BODY_BLOCKED_PARAMETER    = "txn.dd.body"

-- Parameter for routing between nominal backend and failure backend in haproxy configuration
DATADOME_STATUS_VARIABLE = "txn.dd.status"

-- Separator for Headers
HEADER_SEPARATOR = "@@"


-- ///////////////////////////////////////////////
function buildBody(txn, cookieLen)

    local body = {
    ['Key']                = txn:get_var('proc.Datadome_key'),
    ['RequestModuleName']  = DATADOME_MODULE_NAME,
    ['ModuleVersion']      = DATADOME_MODULE_VERSION,
    ['ServerName']         = host,
    ['APIConnectionState'] = helpers.apiConnectionState(txn),
    ['IP']                 = txn.f:src(),
    ['Port']               = port,
    ['TimeRequest']        = helpers.getCurrentMicroTime(),
    ['Protocol']           = helpers.getProtocol(txn),
    ['Method']             = txn.sf:method(),
    ['ServerHostname']     = txn.f:fe_name(), -- TODO - Validate this field
    ['Request']            = txn.f:path(),
    ['HeadersList']        = helpers.buildKeyList(txn.http:req_get_headers()),
    ['Host']               = txn.f:req_fhdr("Host"),
    ['UserAgent']          = txn.sf:req_fhdr("User-Agent"),
    ['Referer']            = txn.sf:req_fhdr("Referer"),
    ['Accept']             = txn.sf:req_fhdr("Accept"),
    ['AcceptEncoding']     = txn.sf:req_fhdr("Accept-encoding"),
    ['AcceptLanguage']     = txn.sf:req_fhdr("Accept-Language"),
    ['AcceptCharset']      = txn.sf:req_fhdr("Accept-Charset"),
    ['Origin']             = txn.sf:req_fhdr("Origin"),
    ['XForwardedForIP']    = txn.sf:req_fhdr("X-Forwarded-For"),
    ['X-Requested-With']   = txn.sf:req_hdr("X-Requested-With"),
    ['Connection']         = txn.sf:req_hdr('connection'),
    ['Pragma']             = txn.sf:req_hdr('pragma'),
    ['CacheControl']       = txn.sf:req_hdr('cache-control'),
    ['ContentType']        = txn.sf:req_hdr('content-type'),
    ['From']               = txn.sf:req_hdr('from'),
    ['X-Real-IP']          = txn.sf:req_hdr('x-real-ip'),
    ['Via']                = txn.sf:req_hdr('via'),
    ['TrueClientIP']       = txn.sf:req_hdr('true-client-ip'),
    ['CookiesLen']         = tostring(cookieLen),
    ['AuthorizationLen']   = tostring(string.len(txn.sf:req_hdr('authorization' or ""))),
    ['PostParamLen']       = txn.sf:req_hdr('content-length'),
    ['SecCHUA']            = txn.sf:req_hdr("Sec-CH-UA"),
    ['SecCHUAArch']        = txn.sf:req_hdr("Sec-CH-UA-Arch"),
    ['SecCHUAModel']       = txn.sf:req_hdr("Sec-CH-UA-Model"),
    ['SecCHUAMobile']      = txn.sf:req_hdr("Sec-CH-UA-Mobile"),
    ['SecCHUAPlatform']    = txn.sf:req_hdr("Sec-CH-UA-Platform"),
    ['SecCHUAFullVersionList'] = txn.sf:req_hdr("Sec-CH-UA-Full-Version-List"),
    ['SecCHDeviceMemory']  = txn.sf:req_hdr("Sec-CH-Device-Memory"),
    ['SecFetchDest']       = txn.sf:req_hdr("Sec-Fetch-Dest"),
    ['SecFetchMode']       = txn.sf:req_hdr("Sec-Fetch-Mode"),
    ['SecFetchSite']       = txn.sf:req_hdr("Sec-Fetch-Site"),
    ['SecFetchUser']       = txn.sf:req_hdr("Sec-Fetch-User"),
    }
    helpers.truncateHeaders(body)
    return body;
end

-- //////
function extractHeaderList(headers, headerListName)
  -- See here, LUA Concat is slow : https://fossies.org/linux/haproxy/doc/lua-api/index.rst
  local result = core.concat()
  if headers[headerListName] ~= nil then
    local listHeaders = helpers.split(headers[headerListName][0], " ")
    for k,v in pairs(listHeaders) do
      result:add(string.lower(v))
      result:add(': ')
      result:add(headers[string.lower(v)][0])
      result:add(HEADER_SEPARATOR)
    end
  end
    return result:dump()
end

core.register_action("Datadome_request_hook", { "http-req" }, function(txn)

	-- Verify that the request is authorized
	local httpclient = core.httpclient()
  local datadomeHeaders = {
		["Connection"] = "keep-alive",
		['Content-Type'] = {"application/x-www-form-urlencoded"},
	}


	local clientId, cookieLen = helpers.getClientIdAndCookiesLength(txn.sf:req_hdr('Cookie'))
	local bodyDD = buildBody(txn, cookieLen);

	bodyDD, datadomeHeaders = helpers.handleClientId(txn, bodyDD, datadomeHeaders, clientId);

	-- Call to the API
	local res = httpclient:post {
		url     = DATADOME_API_CONFIG.api_url,
		headers = datadomeHeaders,
		body    = helpers.stringify(bodyDD),
		timeout = DATADOME_API_CONFIG.timeout, -- in ms
	}
  core.yield() --?

  if ( res.status == 504 ) then -- gateway timeout
    txn:set_var("txn.dd.error", "504 - API Server timed out")
    txn:set_var("txn.request_verified", true)
    return
  else
    if ( res.status == 400 ) then -- Bad request
      txn:set_var("txn.dd.error", "400 - Bad Request " .. helpers.dump(res))
      txn:set_var("txn.request_verified", true)
      return
    end
  end
  if ( res.headers["x-datadomeresponse"] == nil or tonumber(res.headers["x-datadomeresponse"][0]) ~= res.status ) then
    txn:set_var("txn.dd.error", "503 - Invalid X-DataDomeResponse header, is it ApiServer response?")
    return
  end

  -- add header response to variable
  txn:set_var(DATADOME_RESPONSE_HEADERS_PARAMETER, extractHeaderList(res.headers, "x-datadome-headers"))
  -- add header request to variable
  txn:set_var(DATADOME_REQUEST_HEADERS_PARAMETER, extractHeaderList(res.headers, "x-datadome-request-headers"))

	-- Validate the return status
	if ( res.status == 401 or res.status == 403 ) then
		txn:set_var("txn.request_verified", false)
        txn:set_var(DATADOME_STATUS_VARIABLE, "blocked")
		txn:set_var(DATADOME_RESPONSE_CODE_PARAMETER, res.status)
		txn:set_var(DATADOME_RESPONSE_BODY_BLOCKED_PARAMETER, res.body)
	else
		txn:set_var("txn.request_verified", true)
	end

 end)

 -- Add headers to the response to the end-user customer
core.register_action("Datadome_response_hook", { "http-res" }, function(txn)
	local responseHeaders = txn:get_var(DATADOME_RESPONSE_HEADERS_PARAMETER);
  if (responseHeaders ~= nil ) then
      for k,headerNameValue in pairs(helpers.split(responseHeaders,HEADER_SEPARATOR)) do
          local sepPosition = string.find(headerNameValue,": ")
          if (sepPosition ~= nil) then
              local headerName = string.sub(headerNameValue,1,sepPosition-1)
              txn.http:res_add_header(headerName, string.sub(headerNameValue,sepPosition+2));
          end
      end
  end
end)


 -- Blocking page backend
core.register_service("failure_service", "http", function(applet)
	local response = applet:get_var(DATADOME_RESPONSE_BODY_BLOCKED_PARAMETER);
	local response_code = applet:get_var(DATADOME_RESPONSE_CODE_PARAMETER);
	applet:set_status(response_code)
	applet:add_header("content-length", string.len(response))
	applet:add_header("content-type", "text/html")
	applet:start_response()
	applet:send(response)
end)


-- Utils for Logs DataDome Headers
core.register_fetches("ddHeaders", function(txn, value)
  local ddHeaders = txn:get_var(DATADOME_REQUEST_HEADERS_PARAMETER);
  if (ddHeaders ~= nil) then
    for k,headerNameValue in pairs(helpers.split(ddHeaders,HEADER_SEPARATOR)) do
      --core.log(core.info, type(value) .. ' & ' .. value .. ' -k: ' .. k .. ', headerNameValue:' .. headerNameValue .. ' &type:' .. type(headerNameValue) )

      local sepPositionStart, sepPositionEnd = string.find(headerNameValue, string.lower(value), nil, true)

      if (sepPositionStart ~= nil) then
          -- core.log(core.info, 'value: ' .. value .. ', headerNameValue:' .. headerNameValue .. ', seq:' .. dump(sepPositionStart) .. ' end:' .. dump(sepPositionEnd))

          local headerName = string.sub(headerNameValue, sepPositionEnd + 3)
          core.log(core.debug, 'found! ' ..value .. '-' ..headerName)
          return headerName
      end
    end
  end
end)